@@ -4,8 +4,9 @@
 
 from datetime import datetime, timedelta
 from google.transit import gtfs_realtime_pb2
-import requests.exceptions
+
 import requests
+
 import logging
 import re
 import json					#For the output
@@ -30,10 +31,11 @@ ATTR_CHANGES = 'changes'
 
 ATTR_OCCUPANCY = 'occupancy'
 
+ATTR_AVMS_TRIP_ID = 'avms_trip_id'
 ATTR_REAL_TIME_TRIP_ID = 'real_time_trip_id'
 ATTR_LATITUDE = 'latitude'
 ATTR_LONGITUDE = 'longitude'
-
+ATTR_GTFS_URI = 'gtfs_uri'
 ATTR_ALERTS = 'alerts'
 
 logger = logging.getLogger(__name__)
@@ -47,14 +49,6 @@ class TransportNSWv2(object):
 
     def __init__(self):
         """Initialize the data object with default values."""
-        self.origin_id = None
-        self.destination_id = None
-        self.api_key = None
-        self.journey_wait_time = None
-        self.transport_type = None
-        self.strict_transport_type = None
-        self.raw_output = None
-        self.journeys_to_return = None
         self.info = {
             ATTR_DUE_IN : 'n/a',
             ATTR_ORIGIN_STOP_ID : 'n/a',
@@ -76,76 +70,135 @@ class TransportNSWv2(object):
             ATTR_ALERTS: '[]'
             }
 
-    def get_trip(self, name_origin, name_destination , api_key, journey_wait_time = 0, transport_type = 0, \
-                 strict_transport_type = False, raw_output = False, journeys_to_return = 1, route_filter = '', \
-                 include_realtime_location = True, include_alerts = 'none', alert_type = 'all', check_stop_ids = True, forced_gtfs_uri = []):
 
-        """Get the latest data from Transport NSW."""
-        fmt = '%Y-%m-%dT%H:%M:%SZ'
+    def check_stops(self, api_key, stops, home_assistant = False):
+        # Check the list of stops and return a JSON array of the stop details, plus if all the checked stops existed
+        # Return a JSON array of the results
 
-        self.name_origin = name_origin
-        self.destination = name_destination
-        self.api_key = api_key
-        self.journey_wait_time = journey_wait_time
-        self.transport_type = transport_type
-        self.strict_transport_type = strict_transport_type
-        self.raw_output = raw_output
-        self.journeys_to_return = journeys_to_return
-        self.route_filter = route_filter.lower()
-        self.include_realtime_location = include_realtime_location
-        self.include_alerts = include_alerts.lower()
-        self.alert_type = alert_type.lower()
+        # Sanity checking
+        if isinstance(stops, str):
+            # If it's a single string, convert it to a list
+            stops = [stops]
 
-        # This query always uses the current date and time - but add in any 'journey_wait_time' minutes
-        now_plus_wait = datetime.now() + timedelta(minutes = journey_wait_time)
-        itdDate = now_plus_wait.strftime('%Y%m%d')
-        itdTime = now_plus_wait.strftime('%H%M')
-
-        auth = 'apikey ' + self.api_key
+        auth = 'apikey ' + api_key
         header = {'Accept': 'application/json', 'Authorization': auth}
 
+        #Prepare the output string
+        all_stops_valid = True
+        stop_list = []
+        skip_api_calls = False
 
-        # First, check if the source and dest stops are valid unless we've been told not to
-        if check_stop_ids:
-            stop_list = [name_origin, name_destination]
-
-            for stop in stop_list:
+        try:
+            for stop in stops:
                 url = \
                     'https://api.transport.nsw.gov.au/v1/tp/stop_finder?' \
                     'outputFormat=rapidJSON&coordOutputFormat=EPSG%3A4326' \
                     '&type_sf=stop&name_sf=' + stop + \
                     '&TfNSWSF=true'
 
-                # Send the query and return an error if something goes wrong
-                try:
+                # Send the query
+                url = 'https://api.transport.nsw.gov.au/v1/tp/stop_finder?outputFormat=rapidJSON&coordOutputFormat=EPSG%3A4326&type_sf=stop&name_sf=' + stop + '&TfNSWSF=true'
+                error_code = 0
+
+                if not skip_api_calls:
                     response = requests.get(url, headers=header, timeout=5)
+                else:
+                    # An earlier call resulted in an API key error so no point trying again
+                    response.status_code = 401
+
+                # If we get bad status code, handle it depending on the error type
+                if response.status_code != 200:
+                    # We can't be sure that all the stops are valid
+                    error_code = response.status_code
+                    stop_detail = []
+
+                    if response.status_code == 401:
+                        raise InvalidAPIKey("Invalid API key")
+
+                    elif response.status_code == 403 or response.status_code == 429:
+                        raise APIRateLimitExceeded("API rate limit exceeded")
 
-                    # If we get bad status code, log error and return with None
-                    if response.status_code != 200:
-                        if response.status_code == 429:
-                            # We've exceeded the rate limit but that doesn't mean future calls won't work
-                            # So let's assume that the stop ID is ok but we'll still raise a warning
-                            logger.warn(f"Error {str(response.status_code)} calling /v1/tp/stop_finder API; rate limit exceeded - assuming stop is valid")
-                        else:
-                            # If it's an API key issue there's no point in continuing, hence returning None
-                            logger.error(f"Error {str(response.status_code)} calling /v1/tp/stop_finder API; check API key")
-                            return None
                     else:
-                        # Parse the result as a JSON object
-                        result = response.json()
+                        raise StopError("Unknown")
 
-                        # Just a quick check - the presence of systemMessages signifies an error, otherwise we assume it's ok
-                        if 'systemMessages' in result:
-                            logger.error(f"Error - Stop ID {stop} doesn't exist")
-                            return None
+                else:
+                    # Parse the result as a JSON object
+                    stop_detail = response.json()
+                    # Just a quick check - the presence of systemMessages signifies an error, otherwise we assume it's ok
+                    if 'systemMessages' in stop_detail:
+                        error_text = stop_detail['systemMessages'][0]['text']
+                        raise StopError(f"{error_text}", stop)
 
-                        # Put in a pause here to try and make sure we stay under the 5 API calls/second limit
-                        # Not usually an issue but if multiple processes are running multiple calls we might hit it
-                        time.sleep(1.0)
+                    # Put in a pause here to try and make sure we stay under the 5 API calls/second limit
+                    # Not usually an issue but if multiple processes are running multiple calls we might hit it
+                    time.sleep(1.0)
 
-                except Exception as ex:
-                    logger.error(f"Error {str(ex)} calling /v1/tp/stop_finder API; assuming stop is valid")
-                    return None
+                # Append the results to the JSON output - only return the 'isBest' location entry if there's more than one
+                if stop_detail != []:
+                    stop_valid = True
+                    for location in stop_detail['locations']:
+                        if location['isBest']:
+                            stop_detail = location
+                            break
+
+                else:
+                    stop_valid = False
+                    all_stops_valid = False
+
+                #Add it to the list
+                data = {"stop_id": stop, "valid": stop_valid, "error_code": error_code, "stop_detail": stop_detail}
+                stop_list.append (data)
+
+            # Complete the JSON output and return it
+            data = {"all_stops_valid": all_stops_valid, "stop_list": stop_list}
+
+            return data
+
+        except InvalidAPIKey as ex:
+            raise InvalidAPIKey (f"Invalid API key {api_key}")
+
+        except StopError as ex:
+            raise StopError (f"Error '{ex}' calling stop finder API for stop ID {stop}")
+
+        except Exception as ex:
+            raise StopError(f"Error '{ex}' calling stop finder API for stop ID {stop}", stop)
+
+
+    def get_trip(self, name_origin, name_destination , api_key, journey_wait_time = 0, transport_type = 0, \
+                 strict_transport_type = False, raw_output = False, journeys_to_return = 1, route_filter = '', \
+                 include_realtime_location = True, include_alerts = 'none', alert_type = 'all', check_stop_ids = True, forced_gtfs_uri = [],
+                 home_assistant = False):
+
+        """Get the latest data from Transport NSW."""
+        fmt = '%Y-%m-%dT%H:%M:%SZ'
+
+        route_filter = route_filter.lower()
+        include_alerts = include_alerts.lower()
+        alert_type = alert_type.lower()
+
+        # This query always uses the current date and time - but add in any 'journey_wait_time' minutes
+        now_plus_wait = datetime.now() + timedelta(minutes = journey_wait_time)
+        itdDate = now_plus_wait.strftime('%Y%m%d')
+        itdTime = now_plus_wait.strftime('%H%M')
+
+        auth = 'apikey ' + api_key
+        header = {'Accept': 'application/json', 'Authorization': auth}
+
+        # First, check if the source and dest stops are valid unless we've been told not to
+        if check_stop_ids:
+            stop_list = [name_origin, name_destination]
+            data = self.check_stops(api_key, stop_list)
+
+            if not data['all_stops_valid']:
+                # One or both of those stops was invalid - log an error and exit
+                stop_error = ""
+                for stop in data['stop_list']:
+                    if not stop['valid']:
+                        stop_error += stop['stop_id']+ ", "
+
+                #logger.error(f"Stop ID(s) {stop_error[:-2]} do not exist - exiting")
+                raise StopError (f"Stop ID(s) {stop_error[:-2]} do not exist", stop_error)
+                #return None
 
         # We don't control how many journeys are returned any more, so need to be careful of running out of valid journeys if there is a filter in place, particularly a strict filter
         # It would be more efficient to return one journey, check if the filter is met and then retrieve the next one via a new query if not, but for now we'll only be making use of the journeys we've been given
@@ -155,29 +208,32 @@ class TransportNSWv2(object):
             'https://api.transport.nsw.gov.au/v1/tp/trip?' \
             'outputFormat=rapidJSON&coordOutputFormat=EPSG%3A4326' \
             '&depArrMacro=dep&itdDate=' + itdDate + '&itdTime=' + itdTime + \
-            '&type_origin=any&name_origin=' + self.name_origin + \
-            '&type_destination=any&name_destination=' + self.destination + \
+            '&type_origin=any&name_origin=' + name_origin + \
+            '&type_destination=any&name_destination=' + name_destination + \
             '&TfNSWTR=true'
-            # '&calcNumberOfTrips=' + str(journeys_to_retrieve) + \
 
         # Send the query and return an error if something goes wrong
         # Otherwise store the response for the next steps
         try:
             response = requests.get(url, headers=header, timeout=10)
+
         except Exception as ex:
-            logger.error(f"Error {str(ex)} calling /v1/tp/trip API")
-            return None
+            raise TripError (f"Error '{str(ex)}' calling trip API for journey {name_origin} to {name_destination}")
 
         # If we get bad status code, log error and return with n/a or an empty string
         if response.status_code != 200:
-           if response.status_code == 429:
-               logger.error(f"Error {str(response.status_code)} calling /v1/tp/trip API; rate limit exceeded")
-           else:
-               logger.error(f"Error {str(response.status_code)} calling /v1/tp/trip API; check API key")
+            if response.status_code == 401:
+                # API key issue
+                raise InvalidAPIKey("Error 'Invalid API key' calling trip API for journey {name_origin} to {name_destination}")
 
-           return None
+            elif response.status_code == 403 or response.status_code == 429:
+                raise APIRateLimitExceeded("Error 'API rate limit exceeded' calling trip API for journey {name_origin} to {name_destination}")
+
+            else:
+                raise TripError(f"Error '{str(response.status_cude)}' calling trip API for journey {name_origin} to {name_destination}")
+
+            return None
 
-        # Parse the result as a JSON object
         result = response.json()
 
         # The API will always return a valid trip, so it's just a case of grabbing the metadata that we need...
@@ -195,10 +251,7 @@ class TransportNSWv2(object):
             retrieved_journeys = len(result['journeys'])
 
         except:
-            # Looks like an empty response
-            logger.error(f"Error {(str(err))} calling /v1/tp/trip API")
-
-            return None
+            raise TripError(f"Error 'no journeys returned' calling trip API for journey {name_origin} to {name_destination}")
 
         # Loop through the results applying filters where required, and generate the appropriate JSON output including an array of in-scope trips
         json_output=''
@@ -207,17 +260,17 @@ class TransportNSWv2(object):
 
         for current_journey_index in range (0, retrieved_journeys, 1):
             # Look for a trip with a matching transport type filter in at least one of its legs.  Either ANY, or the first leg, depending on how strict we're being
-            journey, next_journey_index = self.find_next_journey(result['journeys'], current_journey_index, transport_type, strict_transport_type, route_filter)
+            journey, next_journey_index = self._find_next_journey(result['journeys'], current_journey_index, transport_type, strict_transport_type, route_filter)
 
             if ((journey is None) or (journey['legs']) is None):
                 pass
             else:
                 legs = journey['legs']
-                first_leg = self.find_first_leg(legs, transport_type, strict_transport_type, route_filter)
+                first_leg = self._find_first_leg(legs, transport_type, strict_transport_type, route_filter)
 
                 #Executive decision - don't be strict on the last leg, there's often some walking (transport type 100) involved.
-                last_leg = self.find_last_leg(legs, transport_type, False)
-                changes = self.find_changes(legs, transport_type)
+                last_leg = self._find_last_leg(legs, transport_type, False)
+                changes = self._find_changes(legs, transport_type)
 
                 origin = first_leg['origin']
                 first_stop = first_leg['destination']
@@ -235,7 +288,7 @@ class TransportNSWv2(object):
                 delay = int((t1-t2) / 60)
 
                 # How long until it leaves?
-                due = self.get_due(datetime.strptime(origin_departure_time, fmt))
+                due = self._get_due(datetime.strptime(origin_departure_time, fmt))
 
                 # Destination info
                 destination_stop_id = destination['id']
@@ -243,7 +296,7 @@ class TransportNSWv2(object):
                 destination_arrival_time = destination['arrivalTimeEstimated']
 
                 # Origin type info - train, bus, etc
-                origin_mode = self.get_mode(transportation['product']['class'])
+                origin_mode = self._get_mode(transportation['product']['class'])
                 origin_mode_name = transportation['product']['name']
 
                 # RealTimeTripID info so we can try and get the current location later
@@ -254,6 +307,11 @@ class TransportNSWv2(object):
                     # We're also going to need the agency_id if it's a bus journey
                     agencyid = transportation['operator']['id']
 
+                # AVMSTripID is for Home Assistant, if needed
+                avmstripid = 'n/a'
+                if 'properties' in transportation and 'AVMSTripID' in transportation['properties']:
+                    avmstripid = transportation['properties']['AVMSTripID']
+
                 # Line info
                 origin_line_name_short = "unknown"
                 if 'disassembledName' in transportation:
@@ -269,26 +327,27 @@ class TransportNSWv2(object):
                     occupancy = first_stop['properties']['occupancy']
 
                 alerts = "[]"
-                if self.include_alerts != 'none':
+                if include_alerts != 'none':
                     # We'll be adding these to the returned JSON string as an array
                     # Only include alerts of the specified priority or greater, and of the specified type
-                    alerts = self.find_alerts(legs, self.include_alerts, self.alert_type)
+                    alerts = self._find_alerts(legs, include_alerts, alert_type)
 
                 latitude = 'n/a'
                 longitude = 'n/a'
+                mode_url = 'n/a'
 
-                if self.include_realtime_location and realtimetripid != 'n/a':
+                if include_realtime_location and realtimetripid != 'n/a':
                     # See if we can get the latitute and longitude via the Realtime Vehicle Positions API
                     # Build the URL(s) - some modes have multiple GTFS sources, unforunately
                     # Some travel modes require brute-forcing the API call a few times, so if we're sure of the URI,
                     # ie it's been determined elsewhere then it can be forced
 
                     bFoundTripID = False
-                    url_base_path = self.get_base_url(origin_mode)
+                    url_base_path = self._get_base_url(origin_mode)
 
                     # Check for a forced URI
                     if not forced_gtfs_uri:
-                        url_mode_list = self.get_mode_list(origin_mode, agencyid)
+                        url_mode_list = self._get_mode_list(origin_mode, agencyid)
                     else:
                         # We've been forced to use a specific URI!
                         url_mode_list = forced_gtfs_uri
@@ -316,10 +375,12 @@ class TransportNSWv2(object):
                                         break
                             else:
                                 # Warn that we didn't get a good return
-                                if response.status_code == 429:
-                                    logger.error(f"Error {str(response.status_code)} calling {url} API; rate limit exceeded")
+                                if response.status_code == 401:
+                                    logger.error(f"Error 'Invalid API key' calling {url} API")
+                                elif response.status_code == 403 or response.status_code == 429:
+                                    logger.error(f"Error 'API rate limit exceded' calling {url} API")
                                 else:
-                                    logger.error(f"Error {str(response.status_code)} calling {url} API; check API key")
+                                    logger.error(f"Error '{str(response.status_code)}' calling {url} API")
 
                             if bFoundTripID == True:
                                 # No need to look any further
@@ -345,11 +406,17 @@ class TransportNSWv2(object):
                     ATTR_CHANGES: changes,
                     ATTR_OCCUPANCY : occupancy,
                     ATTR_REAL_TIME_TRIP_ID : realtimetripid,
+                    ATTR_GTFS_URI: mode_url,
                     ATTR_LATITUDE : latitude,
                     ATTR_LONGITUDE : longitude,
                     ATTR_ALERTS: json.loads(alerts)
                     }
 
+                if home_assistant:
+                    self.info.update (
+                        {ATTR_AVMS_TRIP_ID: avmstripid}
+                    )
+
                 found_journeys = found_journeys + 1
 
                 # Add to the return array
@@ -366,12 +433,12 @@ class TransportNSWv2(object):
 
                 current_journey_index = next_journey_index
 
-        json_output='{"journeys_to_return": ' + str(self.journeys_to_return) + ', "journeys_with_data": ' + str(found_journeys) + ', "journeys": [' + json_output + ']}'
+        json_output='{"journeys_to_return": ' + str(journeys_to_return) + ', "journeys_with_data": ' + str(found_journeys) + ', "journeys": [' + json_output + ']}'
         return json_output
 
 
-    def find_next_journey(self, journeys, start_journey_index, journeytype, strict, route_filter):
-        # Fnd the next journey that has a leg of the requested type, and/or that satisfies the route filter
+    def _find_next_journey(self, journeys, start_journey_index, journeytype, strict, route_filter):
+        # Find the next journey that has a leg of the requested type, and/or that satisfies the route filter
         journey_count = len(journeys)
 
         # Some basic error checking
@@ -379,7 +446,7 @@ class TransportNSWv2(object):
             return None, None
 
         for journey_index in range (start_journey_index, journey_count, 1):
-            leg = self.find_first_leg(journeys[journey_index]['legs'], journeytype, strict, route_filter)
+            leg = self._find_first_leg(journeys[journey_index]['legs'], journeytype, strict, route_filter)
             if leg is not None:
                 return journeys[journey_index], journey_index + 1
             else:
@@ -389,7 +456,7 @@ class TransportNSWv2(object):
         return None, None
 
 
-    def find_first_leg(self, legs, legtype, strict, route_filter):
+    def _find_first_leg(self, legs, legtype, strict, route_filter):
         # Find the first leg of the requested type
         leg_count = len(legs)
         for leg_index in range (0, leg_count, 1):
@@ -419,7 +486,7 @@ class TransportNSWv2(object):
         return None
 
 
-    def find_last_leg(self, legs, legtype, strict):
+    def _find_last_leg(self, legs, legtype, strict):
         # Find the last leg of the requested type
         leg_count = len(legs)
         for leg_index in range (leg_count - 1, -1, -1):
@@ -441,7 +508,7 @@ class TransportNSWv2(object):
         return None
 
 
-    def find_changes(self, legs, legtype):
+    def _find_changes(self, legs, legtype):
         # Find out how often we have to change
         changes = 0
         leg_count = len(legs)
@@ -454,11 +521,11 @@ class TransportNSWv2(object):
         return changes - 1
 
 
-    def find_alerts(self, legs, priority_filter, alert_type):
+    def _find_alerts(self, legs, priority_filter, alert_type):
         # Return an array of all the alerts on this trip that meet the priority level and alert type
         leg_count = len(legs)
         found_alerts = []
-        priority_minimum = self.get_alert_priority(priority_filter)
+        priority_minimum = self._get_alert_priority(priority_filter)
         alert_list = alert_type.split("|")
 
         for leg_index in range (0, leg_count, 1):
@@ -466,14 +533,14 @@ class TransportNSWv2(object):
             if 'infos' in current_leg:
                 alerts = current_leg['infos']
                 for alert in alerts:
-                    if (self.get_alert_priority(alert['priority'])) >= priority_minimum:
+                    if (self._get_alert_priority(alert['priority'])) >= priority_minimum:
                         if (alert_type == 'all') or (alert['type'].lower() in alert_list):
                             found_alerts.append (alert)
 
         return json.dumps(found_alerts)
 
 
-    def find_hints(self, legs, legtype, priority):
+    def _find_hints(self, legs, legtype, priority):
         # Return an array of all the hints on this trip that meet the priority type
         leg_count = len(legs)
 
@@ -484,7 +551,7 @@ class TransportNSWv2(object):
                 hints = current_leg['hints']
 
 
-    def get_mode(self, iconId):
+    def _get_mode(self, iconId):
         """Map the iconId to a full text string"""
         modes = {
             1   : "Train",
@@ -501,7 +568,7 @@ class TransportNSWv2(object):
 
         return modes.get(iconId, None)
 
-    def get_base_url(self, mode):
+    def _get_base_url(self, mode):
         # Map the journey mode to the proper base real time location URL
         v1_url = "https://api.transport.nsw.gov.au/v1/gtfs/vehiclepos"
         v2_url = "https://api.transport.nsw.gov.au/v2/gtfs/vehiclepos"
@@ -519,7 +586,7 @@ class TransportNSWv2(object):
         return url_options.get(mode, None)
 
 
-    def get_alert_priority(self, alert_priority):
+    def _get_alert_priority(self, alert_priority):
         # Map the alert priority to a number so we can filter later
 
         alert_priorities = {
@@ -533,7 +600,7 @@ class TransportNSWv2(object):
         return alert_priorities.get(alert_priority.lower(), 4)
 
 
-    def get_mode_list(self, mode, agencyid):
+    def _get_mode_list(self, mode, agencyid):
         """
         Map the journey mode to the proper modifier URL.  If the mode is Bus, Coach or School bus then use the agency ID to invoke the GTFS datastore search API
         which will give us the appropriate URL to call later - we still have to do light rail the old-fashioned, brute-force way though
@@ -548,15 +615,17 @@ class TransportNSWv2(object):
             try:
                 response = requests.get(url, timeout=5)
             except Exception as ex:
-                logger.error("Error " + str(ex) + " querying GTFS URL datastore")
+                logger.error(f"Error '{str(ex)}' querying GTFS URL datastore")
                 return None
 
             # If we get bad status code, log error and return with None
             if response.status_code != 200:
-                if response.status_code == 429:
-                    logger.error("Error " + str(response.status_code) + " calling /v1/tp/stop_finder API; rate limit exceeded")
+                if response.status_code == 401:
+                    logger.error (f"Error 'Invalid API key' calling GTFS API url {url}")
+                elif response.status_code == 403 or response.status_code == 429:
+                    logger.error(f"Error 'API rate limit exceeded' calling GTFS API url {url}")
                 else:
-                    logger.error("Error " + str(response.status_code) + " calling /v1/tp/stop_finder API; check API key")
+                    logger.error(f"Error '{str(response.status_code)}' calling GTFS API url {url}")
 
                 return None
 
@@ -581,9 +650,29 @@ class TransportNSWv2(object):
             return url_options.get(mode, None)
 
 
-    def get_due(self, estimated):
+    def _get_due(self, estimated):
         # Minutes until departure
         due = 0
         if estimated > datetime.utcnow():
             due = round((estimated - datetime.utcnow()).seconds / 60)
         return due
+
+# Exceptions
+class InvalidAPIKey(Exception):
+    """ API key error """
+
+class APIRateLimitExceeded(Exception):
+    """ API rate limit exceeded """
+
+class StopError(Exception):
+    """ Stop-finder related error """
+    def __init__(self, message = "", stop_detail = ""):
+        super().__init__(message)
+        self.stop_detail = stop_detail
+
+#    def __str__(self):
+#        return f"{self.message}"
+
+class TripError(Exception):
+    """" Trip-finder related error """
+
